// File: BoSL_Radar.ino

/************************************************/
/************ Connections for setup **************
* Radar VPP (Green wire)       -- SWB3V3
* Radar GND (Green-white wire) -- GND
* Radar TX (Blue wire)         -- TX2
* Radar RX (Brown wire)        -- RX2
* Radar RESET (Orange wire)    -- D13
*************************************************/

/* this code requires arduino IDE > 2.3.2  */

// ********* Configurations for setup ************
#define SITE_DIR "testing"
#define SITE_ID "B5_RADAR"
#define APN "simbase"
#define MCCMNC "50501"
#define SD_LOG_ALL_DATA      true       // save raw sensor readings to SD card.
#define SD_LOG_FILE_ALL_DATA "log.txt"  // file to save raw sensor readings (max 8.3 characters)
#define SD_LOG_FILE_WEB_DATA  "web.txt" // file to save averaged sensor readings (max 8.3 characters)
#define LOG_TO_WEB true                 // logging to web only occurs when true. if false SD log times may not be accurate
#define ALLOW_GSM false                 // most networks have disabled GSM
#define ALLOW_NBIOT false               // NB-IoT requires simcard support
#define ACCEPT_NON_ROAMING true         // set false to only accept roaming connections
#define LOOP_INTERVAL_MINUTES 1         // sensor scan interval (minutes)
#define AVERAGING_INTERVAL_MINUTES 6    // averaging and upload interval (minutes)
#define CONSTANT_LOGGING true           // if true, it will log to web every
                                        // AVERAGING_INTERVAL_MINUTES

// * Variable logging control * // only used when CONSTANT_LOGGING is false
#define MAX_HTTP_INTERVAL_MINUTES 60 // maximum upload interval (minutes)
/*  In variable logging, a log is triggered when any of the sensor readings change by 
 *  more than an absolute or relative threshold from the previous upload.
 *  The below define these absolute and relative thresholds.
 */
#define THRESHOLD_PERCENT 30.0   // relative threshold for sensor readings (%)
#define VEL_DIFF_THRESHOLD 10.0  // absolute threshold for velocity readings (mm/s)
#define AVEL_DIFF_THRESHOLD 10.0  // absolute threshold for angle corrected velocity readings (mm/s)
#define DIST_DIFF_THRESHOLD 10.0 // absolute threshold for distance readings (mm)

// For Radar
#define LOGMODE 5 // 5: velocity + fine depth. 7: only fine depth (but this can measure further from the sensor).
#define SWEEP_START 500 // (mm) line of sight distance from the radar sensor at which measurement will start
                        // too short a value will lead to interference in the nearest measurements.
                        // too long a value and data will be missed when the water level rises close to the sensor 
                        // if LOGMODE is 5:
                        //    SWEEP_START should be set to > 500 for best results
                        // if LOGMODE is 7:
                        //    SWEEP_START should be set to > 300 for best results (300 if unsure)

#define SWEEP_LENGTH 2000 // (mm) line of sight length along which the radar sensor will perform measurements
                          // longer measurements return more data but will also take longer.
                          // if LOGMODE is 5:
                          //    SWEEP_LENGTH should be set to < 3810.
                          // if LOGMODE is 7:
                          //    SWEEP_LENGTH should be set to < 5300. (5300 if unsure)

//  NOTE: SWEEP_START+SWEEP_LENGTH should always be less than 7000. Ideally, they should be no longer than needed for the measurement range to read the base of the water channel (keeping in mind that line of sight is not the same as vertical height)


//************************************************
//************************************************

// ********* Configurations for error handling ************
#define HTTP_FAILS_RESET_THRESHOLD       (240) //how many HTTP fails are needed to trigger a sim7000 reset
#define TRY_POWER_ON_MAX_RESETS          (3)   //max number of reset attempts when attempting to power on the sim7000

#define POWER_ON_FAILS_RESET_LIMIT_BEGIN (6)   //how many power on resets are needed to rate limit the reset attempts
#define POWER_ON_FAILS_RESET_LIMIT_RATE  (30)  //in the rate limit mode, this is the number of reset requests needed before a reset is actually performed.

#define NET_REG_FAILS_LIMIT_BEGIN (12)   //how many power on resets are needed to rate limit the reset attempts
#define NET_REG_FAILS_LIMIT_RATE  (120)  //in the rate limit mode, this is the number of reset requests needed before a reset is actually performed.

//*************************************************************
// Do NOT touch bellow code unless you know what you are doing.
//*************************************************************

#include <MCP7940.h>
#include <SD.h>
#include <avr/power.h>
#include <avr/wdt.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <time.h>

// *** special macro ***
#define MAX_OPERATORS (8)
#define POWER_ON_TRY_TIMES 5
#define HTTP_UPLOAD_TRY_TIMES 3 // to fix the issues of simcom or bad signal.
// first time try power off simcom, second time try reset simcom, third time
// and forward only try power off simcom to avoid frequent reset

// *** Pins ***
// For the SIM7000
// TODO: use Pin names on the BoSL board instead of numbers
#define PWRKEY 38
#define SIM_RESET 40      // PD6
#define SIM_BUF_EN 25     // PG4
#define SIM_STATUS_BFD 45 // PC1
#define BOSL_TX 36 // TX1
#define BOSL_RX 37 // RX1
// For the radar sensor
#define Radar_VPP 66
#define Radar_TX 9 // TX2/PH1
#define Radar_RX 8 // RX2/PH0
#define RADAR_RST 13 // D13/PH5
// SD card
#define SD_CS 15
// For the RTC
#define RTC_INTERRUPT 7

// *** Serials ***
#define simCom Serial1
#define velCom Serial2

// *** Constants ***
#define LOOP_INTERVAL_MILLIS (LOOP_INTERVAL_MINUTES * 60000) // sensor scan
                                                           // interval(ms)
#define BAUDRATE 9600 // talk to SIM7000 baud rate
#define CHARBUFF 254  // SIM7000 serial response buffer,longer than 255 will cause issues
#define SIM_RESET_MIN_INTERVAL_HOURS 3
#define MIN_NON_RESET_LOOP_COUNT (SIM_RESET_MIN_INTERVAL_HOURS * 60 / LOOP_INTERVAL_MINUTES)

// *** global variables ***
// atmega timer0_millis
extern volatile unsigned long timer0_millis;
// RTC
MCP7940_Class MCP7940;
struct DATETIME {
public:
  init(int16_t yr, int16_t mt, int16_t dy, int16_t hr, int16_t mn, int16_t sc) {
    this->yr = yr;
    this->mt = mt;
    this->dy = dy;
    this->hr = hr;
    this->mn = mn;
    this->sc = sc;
  }
  int16_t yr;
  int16_t mt;
  int16_t dy;
  int16_t hr;
  int16_t mn;
  int16_t sc;
} g_datetime;
// SIM7000
char response[CHARBUFF]; // sim7000 serial response buffer
String dataStr;          // Transmit URL
char CBC[5];
char CSQ[5];
// sensor variables
char data[256];
char VEL[8];
char DIST[8];
char AMP[8];
char SPREAD[8];
char ANGLE[8];
char PEAKS[40];
char FDIST[8];
char AVEL[16];
char AFDIST[16];
// SD card variables
File myFile;
unsigned long sdRecordIndex = 0;
unsigned long webRecordIndex = 0;
bool isSDSetupDone = false;
// SIM RESET control
uint32_t noResetLoopsCount = MIN_NON_RESET_LOOP_COUNT;

// *** logging control ***
int32_t velSum = 0;
int32_t avelSum = 0;
int32_t distSum = 0;
uint8_t valid_scan_count = 0;
int16_t velAverage = 0;
int16_t avelAverage = 0;
int16_t distAverage = 0;
int16_t preVelAverage = 0;
int16_t preAVelAverage = 0;
int16_t preDistAverage = 0;
uint8_t scanCounterForHttp = 0;
uint32_t noLoggingLoopCounter = 0;

// *** function prototypes ***
bool negoSimComBaudRate(int32_t baud = BAUDRATE);
void velread(uint8_t logmode, uint16_t sweep_start = SWEEP_START,
             uint16_t sweep_length = SWEEP_LENGTH);
bool tryPowerOnSimCom(int8_t tries = POWER_ON_TRY_TIMES);
bool sendATcmd(String ATcommand, char *expctAns, uint32_t timeout = 1000,
               int8_t tries = 5);

struct Operator {
    uint8_t status;
    char mccmnc[7];
    uint8_t netact;
    uint8_t rssi;
};

uint8_t operators_len = 0;
Operator operators[MAX_OPERATORS];


// *** function definitions ***
bool serialBegin() {
  pinMode(BOSL_RX, OUTPUT);
  digitalWrite(BOSL_RX, HIGH);
  pinMode(BOSL_TX, INPUT_PULLUP);
  simCom.begin(BAUDRATE);
  xDelay(1000);
  return negoSimComBaudRate(BAUDRATE);
}

bool negoSimComBaudRate(int32_t baud) {
  Serial.println(F("\n== nego baud rate =="));
  simCom.begin(baud);
  if (sendATcmd(F("AT"), "OK", 1000, 16) != 0) {
    Serial.println(F("\nSim7000 current baud rate is 9600 :)"));
    return sendATcmd(F("ATE0"), "OK", 1000, 4);
  }
  return false;
}

void serialEnd() {
  while (simCom.available()) {
    simCom.read();
  }
  simCom.flush();
  xDelay(300);
  simCom.end();
  // pull down RX TX to save power
  pinMode(BOSL_RX, OUTPUT);
  pinMode(BOSL_TX, OUTPUT);
  digitalWrite(BOSL_RX, LOW);
  digitalWrite(BOSL_TX, LOW);
}

bool isSimComOn() { return digitalRead(SIM_STATUS_BFD); }

bool powerOnSimCom() {
  Serial.println(F("\nPowerOnSimCom"));
  delay(50);
  digitalWrite(PWRKEY, LOW);
  xDelay(1000);
  digitalWrite(PWRKEY, HIGH);
  xDelay(4000);
  for (int8_t i = 0; i < 10; i++) {
    if (isSimComOn()) {
      return true;
    }
    xDelay(500);
  }
  return false;
}

bool tryPowerOnSimCom(int8_t tries) {
  Serial.println(F("\nTry to power on SimCom:"));
  if (isSimComOn()) {
    Serial.println(F("\nSimCom is already on"));
    Serial.println(F("succeed."));
    return true;
  }

  for (int8_t i = 0; i < 3; i++) {
    if (powerOnSimCom()) {
      Serial.println(F("succeed."));
      return true;
    }
  }

  return false;
}

bool tryPowerOffSimCom() {
  Serial.println(F("\nTry to power off SimCom:"));
  if (isSimComOn() == false) {
    return true;
  }

  for (int8_t i = 0; i < 3; i++) {
    if (hardPowerOffSimCom()) {
      Serial.println(F("Power off Sim succeed."));
      return true;
    }
  }
  return false;
}

bool hardPowerOffSimCom() {
  Serial.println(F("\nhardPowerOffSimCom"));
  delay(50);
  digitalWrite(PWRKEY, LOW);
  xDelay(1300);
  digitalWrite(PWRKEY, HIGH);
  int8_t i = 0;
  for (; i < 20; i++) {
    if (isSimComOn() == false) {
      return true;
    }
    xDelay(500);
  }
  return false;
}

bool hardResetSimCom() {
  Serial.println(F("Reset SIM7000"));
  pinMode(SIM_RESET, OUTPUT);
  digitalWrite(SIM_RESET, LOW);
  xDelay(2000);
  digitalWrite(SIM_RESET, HIGH);
  xDelay(4000);
  for (int8_t i = 0; i < 10; i++) {
    if (isSimComOn()) {
      return true;
    }
    xDelay(500);
  }
  return false;
}

void openbearer() {
  // set CSTT - if it is already set, then no need to do again...
  sendATcmd(F("AT+CSTT?"), "OK", 1000);
  if (strstr(response, APN) != NULL) {
    // this means the cstt has been set, so no need to set again!
    Serial.println(F("CSTT already set to APN ...no need to set again"));
  } else {
    sendATcmd(F("AT+CSTT=\"" APN "\""), "OK", 1000);
  }
  // close open bearer
  sendATcmd(F("AT+SAPBR=2,1"), "OK", 1000);
  if (strstr(response, "1,1") == NULL) {
    if (strstr(response, "1,3") == NULL) {
      sendATcmd(F("AT+SAPBR=0,1"), "OK", 1000);
    }
    sendATcmd(F("AT+SAPBR=3,1,\"APN\",\"" APN "\""), "OK", 1000);  // set bearer apn
    sendATcmd(F("AT+SAPBR=1,1"), "OK", 1000);
  }
}

void closeBearer() { sendATcmd(F("AT+SAPBR=0,1"), "OK", 1000); }

bool isValidDateTime(const DATETIME &dt) {
  int8_t dtyr = dt.yr % 100;
  if (dt.mt < 0 || dt.mt > 12) {
    return false;
  }
  if (dt.dy < 0 || dt.dy > 31) {
    return false;
  }
  if (dt.hr < 0 || dt.hr > 24) {
    return false;
  }
  if (dt.mn < 0 || dt.mn > 60) {
    return false;
  }
  if (dt.sc < 0 || dt.sc > 60) {
    return false;
  }
  return true;
}

void datetime2y2k(const DATETIME* cal, const int16_t tz4, uint32_t* y2ksecs){
    struct tm t;
    t.tm_sec = cal->sc; 
    t.tm_min = cal->mn; 
    t.tm_hour = cal->hr; 
    t.tm_mday = cal->dy; 
    t.tm_mon = (cal->mt -1); 
    t.tm_year = cal->yr + 100; 
    t.tm_isdst = 0;

    *y2ksecs = mktime(&t);
    *y2ksecs -= 900*(uint32_t)tz4;//we need to subtract the timezone given in quarter hour increments (900 seconds)

}

void y2k2datetime(DATETIME* cal, const uint32_t y2ksecs){
  uint32_t seconds, minutes, hours, days, year, month;
  uint32_t day_of_week;
  seconds = y2ksecs;

  minutes  = seconds / 60;
  hours    = minutes / 60;
  days     = hours   / 24;

  seconds = seconds % 60;
  minutes = minutes % 60;
  hours   = hours   % 24;

  /* avr time starts in 2000 */
  year    = 2000;

  while(1)
  {
    int     leap_year   = (year % 4 == 0 && (year % 100 != 0 || year % 400 == 0));
    uint16_t days_in_year = leap_year ? 366 : 365;
    if (days >= days_in_year)
    {
      days      -= days_in_year;
      ++year;
    }
    else
    {
      /* calculate the month and day */
      static const uint8_t days_in_month[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
      for(month = 0; month < 12; ++month)
      {
        uint8_t dim = days_in_month[month];

        /* add a day to feburary if this is a leap year */
        if (month == 1 && leap_year)
          ++dim;

        if (days >= dim)
          days -= dim;
        else
          break;
      }
      break;
    }
  }

  cal->sc  = seconds;
  cal->mn  = minutes;
  cal->hr = hours;
  cal->dy = days + 1;
  cal->mt  = month + 1;
  cal->yr = year - 2000;//for correct calender epoch
}

bool pullNetTime() {
  int ret = 0;

  if (sendATcmd(F("AT+CLTS=1"), "OK", 1000) == false) {
    return false;
  }
  if (sendATcmd(F("AT+CCLK?"), "OK", 1000) == false) {
    return false;
  }
  char *p = strstr(response, "+CCLK: \"");
  if (p == NULL) {
    return false;
  }

  int16_t tz4;
  ret = sscanf(p, "%s \"%d/%d/%d,%d:%d:%d%d", response, &g_datetime.yr,
         &g_datetime.mt, &g_datetime.dy, &g_datetime.hr, &g_datetime.mn,
         &g_datetime.sc, &tz4);
  if(ret == 8){//number of matches)
    /* fix timezone offset to UTC */
    uint32_t y2ksec;
    datetime2y2k(&g_datetime, tz4, &y2ksec);
    y2k2datetime(&g_datetime, y2ksec);
    return true;
  }
  return false;
}

void _dbgPrintDateTime(const DATETIME& datetime) {
  Serial.print(datetime.yr);
  Serial.print("-");
  Serial.print(datetime.mt);
  Serial.print("-");
  Serial.print(datetime.dy);
  Serial.print(" ");
  Serial.print(datetime.hr);
  Serial.print(":");
  Serial.print(datetime.mn);
  Serial.print(":");
  Serial.print(datetime.sc);
  Serial.println();
}

void initRTC() {
    while (!MCP7940.begin()) {
        Serial.println(F("Unable to find RTC(MCP7940). Checking again in 1000 ms."));
        xDelay(1000);
    }
    Serial.println(F("RTC(MCP7940) initialized."));
    Serial.println(F("Starting RTC..."));
    while (!MCP7940.deviceStatus()){ // Turn oscillator on if necessary
        bool deviceStatus =  MCP7940.deviceStart(); // Start oscillator and return state
        if (!deviceStatus){       // If it didn't start
            xDelay(1000);
        }
    }
    Serial.println(F("RTC started."));

    MCP7940.setBattery(true);     // enable battery backup mode

    Serial.print(F("RTC DateTime:"));
    DateTime now = MCP7940.now();
    char RTCBuffer[30];
    memset(RTCBuffer, '\0', sizeof(RTCBuffer));
    sprintf(RTCBuffer, "%04d-%02d-%02d %02d:%02d:%02d", now.year(), now.month(),
            now.day(), now.hour(), now.minute(), now.second());
    Serial.println(RTCBuffer);
}

void updateRTC(const DATETIME *dt) {
  if (dt == NULL) {
    MCP7940.adjust();
  } else {
    DateTime datetime = DateTime(dt->yr, dt->mt, dt->dy, dt->hr, dt->mn, dt->sc);
    MCP7940.adjust(datetime);
  }
  Serial.print(F("RTC Date/Time update to:"));
  DateTime now = MCP7940.now();
  char RTCBuffer[30];
  memset(RTCBuffer, '\0', sizeof(RTCBuffer));
  sprintf(RTCBuffer, "%04d-%02d-%02d %02d:%02d:%02d", now.year(), now.month(),
          now.day(), now.hour(), now.minute(), now.second());
  Serial.println(RTCBuffer);
}

void afterWakeup() {
  Serial.println(F("\n*** afterWakeup() ***"));
  Serial.flush();
  // detachInterrupt(digitalPinToInterrupt(RTC_INTERRUPT));
  detachInterrupt(RTC_INTERRUPT);
  noSleep();
}

// void SleepySeconds(uint32_t secs) {
//   // set RTC
//   DateTime now = MCP7940.now();
//   now = now + TimeSpan(0, 0, 0, secs);
//   MCP7940.setAlarm(0, matchAll, now, true);
//   //attachInterrupt(digitalPinToInterrupt(RTC_INTERRUPT), afterWakeup, LOW);
//   attachInterrupt(RTC_INTERRUPT, afterWakeup, LOW);

//   sleepMode(SLEEP_POWER_DOWN);
//   // sleepMode(SLEEP_STANDBY);
//   sleep();
// }

// RESET sensors
void resetPin(int8_t pin) {
  pinMode(pin, OUTPUT);
  digitalWrite(pin, LOW);
  delay(10);
  digitalWrite(pin, HIGH);
  pinMode(pin, INPUT);
}

void velread(uint8_t logmode, uint16_t sweep_start, uint16_t sweep_length) {
  Serial.println(F("\nvelread begin."));

  pinMode(Radar_TX, OUTPUT);
  digitalWrite(Radar_TX, HIGH);
  pinMode(Radar_RX, INPUT_PULLUP);
  velCom.begin(9600);

  xDelay(1200);

  resetPin(RADAR_RST);

  while (velCom.available()) {
    velCom.read();
  }

  velCom.setTimeout(2000);
  if (velCom.find("<cfg>", 5)) {
    Serial.println(F("<cfg> received"));
    memset(data, '\0', sizeof(data));
    // *** format ***
    // cfg:{l:5300,b:300,p:0.5,r:500000,d:60,s:64}
    // *** meaning of each parameter ***
    // l: sweep length
    // b: sweep start
    // p: peek threshold
    // r: sweep rate
    // d: distance resolution
    // s: sweeps per measurement
    // *** usage ***
    // not all parameters need to be configured, only the ones that need to be
    sprintf(data, "cfg:{l:%u,b:%u}", SWEEP_LENGTH, SWEEP_START);
    Serial.println(data);
    velCom.println(data);
    velCom.flush();

    velCom.println(F("<debug>"));
    velCom.flush();
    Serial.println(F("<debug> sent"));
  }

  while (velCom.available() && velCom.peek() != '{') {
    velCom.read();
  }

  memset(data, '\0', sizeof(data));
  if (velCom.readBytesUntil('}', data, sizeof(data)) > 0) {
    data[strlen(data)] = '}';
    Serial.println(data);
    Serial.flush();
  }
  memset(data, '\0', sizeof(data));

  delay(10);

  velCom.find("R");

  if (logmode == 1) {
    velCom.write('P');
  } else if (logmode == 3) {
    velCom.write('M');
  } else if (logmode == 4) {
    velCom.write('D');
  } else if (logmode == 5) {
    velCom.write('Q');
  } else if (logmode == 7) {
    velCom.write('E');
  }

  ///timeout
  int variable = 0;
  int isnum = 0;
  int index = 0;

  // // Serial.println(F("velocity summary"));
  velCom.write('A');

  int32_t time = millis();
  uint32_t timeout;

  timeout = 6000;

  memset(VEL, '\0', sizeof(VEL));
  memset(DIST, '\0', sizeof(DIST));
  memset(AMP, '\0', sizeof(AMP));
  memset(SPREAD, '\0', sizeof(SPREAD));
  memset(ANGLE, '\0', sizeof(ANGLE));
  memset(PEAKS, '\0', sizeof(PEAKS));
  memset(FDIST, '\0', sizeof(FDIST));
  memset(AVEL, '\0', sizeof(AVEL));
  memset(AFDIST, '\0', sizeof(AFDIST));

  while (millis() - time < timeout) {
    if (velCom.available()) {
      char chr = velCom.read();
      Serial.print(chr);
      if (chr == '\n') {
        isnum = 0;
      }
      if (chr == '\r') {
        isnum = 0;
      }

      if (isnum == 1) {
        if (!(chr == ' ')) {
          if (logmode != 2) {
            if (logmode != 7) {
              if (variable == 1) { index < (sizeof(VEL)-1) ?  VEL[index] = chr : 0 ; }
              if (variable == 2) { index < (sizeof(DIST)-1) ?  DIST[index] = chr : 0 ; }
              if (variable == 3) { index < (sizeof(AMP)-1) ?  AMP[index] = chr : 0 ; }
              if (variable == 4) { index < (sizeof(SPREAD)-1) ?  SPREAD[index] = chr : 0 ; }
              if (variable == 8) { index < (sizeof(AVEL)-1) ?   AVEL[index] = chr : 0 ; }
            }
            if (variable == 5) { index < (sizeof(ANGLE)-1) ?  ANGLE[index] = chr : 0 ; }
            if (variable == 6) { index < (sizeof(PEAKS)-1) ?  PEAKS[index] = chr : 0 ; }
            if (variable == 7) { index < (sizeof(FDIST)-1) ?  FDIST[index] = chr : 0 ; }
            if (variable == 9) { index < (sizeof(AFDIST)-1) ? AFDIST[index] = chr : 0 ; }
            index++;
          }
        }
      }
      if (chr == ':') {
        variable++;
        index = 0;
        isnum = 1;
      }

      if ((variable == 9) && (isnum == 0) && (index > 1)) { break; }
    }
  }
  Serial.println();
  Serial.println(F("end VEL"));

  Serial.print(F("AVEL:"));
  Serial.println(AVEL);
  Serial.print(F("DIST:"));
  Serial.println(DIST);
  Serial.print(F("ANGLE:"));
  Serial.println(ANGLE);

  // Serial.print(F("AFDIST:"));
  // Serial.println(AFDIST);
  Serial.flush();

  pinMode(Radar_TX, OUTPUT);
  pinMode(Radar_RX, OUTPUT);
  digitalWrite(Radar_TX, LOW);
  digitalWrite(Radar_RX, LOW);

  Serial.println();
  Serial.println(F("velread done."));
}

bool isSDAcessable() { return SD.begin(SD_CS); }

bool SDSetup() {
  Serial.println(F("Initializing SD card..."));
  pinMode(SD_CS, OUTPUT);
  digitalWrite(SD_CS, HIGH);
  if (!SD.begin(SD_CS)) {
    Serial.println(F("Initialization failed!"));
    return false;
  }
  Serial.println(F("Initialization done."));


  DateTime now = MCP7940.now(); // get the current time
  memset(response, '\0', sizeof(response));
  sprintf(response, "%04d-%02d-%02d %02d:%02d:%02d", now.year(), now.month(),
          now.day(), now.hour(), now.minute(), now.second());

  if(SD_LOG_ALL_DATA){
    myFile = SD.open(SD_LOG_FILE_ALL_DATA, FILE_WRITE);
    // if the file opened okay, write to it:
    if (myFile) {
      Serial.println(F("SD card has been set up successfully..."));
      myFile.print(F("Arduino has been reset. New data starts at:"));
      myFile.println(response);
      myFile.println(F("Time,Index,Angle,Velocity,Peaks,Distance,Amplitude,Signal Quality,Angle Corrected Velocity,Angle Corrected Distance"));
      // close the file:
      myFile.close();
      Serial.println(F("SD card and log file set up is Done."));
    } else {
      // if the file didn't open, print an error:
      Serial.print(F("Error opening "));
      Serial.println(SD_LOG_FILE_ALL_DATA);
      return false;
    }
  }

  myFile = SD.open(SD_LOG_FILE_WEB_DATA, FILE_WRITE);
  // if the file opened okay, write to it:
  if (myFile) {
    Serial.println(F("SD card has been set up successfully..."));
    myFile.print(F("Arduino has been reset. New data starts at:"));
    myFile.println(response);
    myFile.println(F("Time,Index,Angle,Velocity,Peaks,Distance,Amplitude,Signal Quality,Angle Corrected Velocity,Angle Corrected Distance"));
    // close the file
    myFile.close();
    Serial.println(F("SD card and log file set up is Done."));
  } else {
    // if the file didn't open, print an error:
    Serial.print(F("Error opening "));
    Serial.println(SD_LOG_FILE_WEB_DATA);
    return false;
  }

  return true;
}

void logToSD(const char* fileName) {
  unsigned long index = 0;
  if (strstr(fileName, SD_LOG_FILE_ALL_DATA)) {
    sdRecordIndex++;
    index = sdRecordIndex;
  } else if (strstr(fileName, SD_LOG_FILE_WEB_DATA)) {
    webRecordIndex++;
    index = webRecordIndex;
  }

  if (isSDAcessable() == false) {
    return;
  }

  myFile = SD.open(fileName, FILE_WRITE);

  if (myFile) {
    DateTime now = MCP7940.now(); // get the current time
    memset(response, '\0', sizeof(response));
    sprintf(response, "%04d-%02d-%02d %02d:%02d:%02d", now.year(), now.month(),
            now.day(), now.hour(), now.minute(), now.second());
    myFile.print(response);
    myFile.print(F(","));
    myFile.print(index);
    myFile.print(F(","));
    myFile.print(ANGLE);
    myFile.print(F(","));
    myFile.print(VEL);
    myFile.print(F(","));
    myFile.print(PEAKS);
    myFile.print(F(","));
    myFile.print(FDIST);
    myFile.print(F(","));
    myFile.println(AMP);
    myFile.print(F(","));
    myFile.println(CSQ);
    myFile.print(F(","));
    myFile.print(AVEL);
    // myFile.print(F(","));
    // myFile.println(AFDIST);
    myFile.close();
    Serial.println(F("Just Saved To SD Card."));

  } else {
    // if the file didn't open, print an error:
    Serial.println(F("error opening datafile"));
    xDelay(1000);
  }
}

bool getLastStatusTimeFromSD() {
  char strtime[32];
  bool ret = false;
  if (isSDAcessable() == false) {
    return false;
  }

  File myFile = SD.open(SD_LOG_FILE_ALL_DATA, FILE_READ);

  if (!myFile) {
    return false;
  }

  unsigned long fileSize = myFile.size();

  if (fileSize < 10) {
    myFile.close();
    return false;
  }

  // seek to the "end" of the file
  if (fileSize > 200) {
    myFile.seek(fileSize - 200);
  }

  // seek to the start of a line
  while (myFile.available()) {
    if (myFile.read() == '\n') {
      break;
    }
  }

  memset(strtime, '\0', sizeof(strtime));
  uint16_t i = 0;
  char c = 0;
  bool commaFound = false;
  while (myFile.available()) {
    c = myFile.read();
    if (c == ',' && commaFound == false) {
      sscanf(strtime, "%d-%d-%d %d:%d:%d", &g_datetime.yr, &g_datetime.mt,
             &g_datetime.dy, &g_datetime.hr, &g_datetime.mn, &g_datetime.sc);
      commaFound = true;
      Serial.println(strtime);
      continue;
    } else if (c == '\n') {
      memset(strtime, '\0', sizeof(strtime));
      i = 0;
      commaFound = false;
      continue;
    } else if (i < sizeof(strtime) && commaFound == false) {
      strtime[i] = c;
      i++;
    }
  }

  myFile.close();

  char buf[20];
  sprintf(buf, "%d-%d-%d %d:%d:%d", g_datetime.yr, g_datetime.mt, g_datetime.dy,
          g_datetime.hr, g_datetime.mn, g_datetime.sc);
  Serial.print(F("Last status time from SD:"));
  Serial.println(buf);

  return isValidDateTime(g_datetime);
}
void _test(){
  Serial.begin(BAUDRATE);


  // _test_netReg();
  // _test_logToWeb();
  _test_printUint16_t();
  
  while(1);
}

int _test_printUint16_t(void){
  char buf[16];

  uint16_t a = 0xFFFF;

  Serial.println("testing print specifiers");

  sprintf(buf, "%" PRIu16, a);
  Serial.println(buf); 

  sprintf(buf, "%u",  a);
  Serial.println(buf); 

  sprintf(buf, "%hu", a);
  Serial.println(buf); 
}

int _test_logToWeb(void){
  for(uint32_t i = 0; i < 0xFFFFFFFF; i++){
    Serial.println("********************************");
    Serial.print(F("loop count: ")); Serial.print(i);
    logToWeb();
    Serial.println("********************************");
    delay(1000);
  }
}

bool logToWeb() {
  bool ret = false;
  static uint16_t upload_fails = 0;
  static uint16_t power_on_fails = 0;
  static uint16_t net_reg_fails = 0;

  Serial.print(F("upload fails: "));
  Serial.print(upload_fails);
  Serial.print(F(" power_on_fails: "));
  Serial.print(power_on_fails);
  Serial.print(F(" net_reg_fails: "));
  Serial.println(net_reg_fails);

  if (net_reg_fails > NET_REG_FAILS_LIMIT_BEGIN && 
      (net_reg_fails % NET_REG_FAILS_LIMIT_RATE)) {
    Serial.println(F("netReg failure high. Limiting logging"));
    net_reg_fails++;
    return false;
  }

  if (net_reg_fails > NET_REG_FAILS_LIMIT_BEGIN &&
     !(net_reg_fails % 3 * NET_REG_FAILS_LIMIT_RATE)) {
    Serial.println(F("netReg failure high. Reseting"));
    hardResetSimCom();
  }

  Serial.println(F("Initialising SIM 7000"));
  simCom.begin(BAUDRATE);

  uint8_t reset = false;
  for (uint8_t i = 0; i < TRY_POWER_ON_MAX_RESETS; i++) {
    if (reset == true) {
      reset = false; 
      power_on_fails++;
      if (power_on_fails < POWER_ON_FAILS_RESET_LIMIT_BEGIN || 
           !(power_on_fails % POWER_ON_FAILS_RESET_LIMIT_RATE)) {
          Serial.println(F("cannot communicate with simCom. resetting"));
          hardResetSimCom();
      }
    }
    
    ret = tryPowerOnSimCom(POWER_ON_TRY_TIMES);
    if(ret == false){reset = true; continue;}

    ret = serialBegin();
    if(ret == false){reset = true; continue;}
    
    break;
  }
  if(ret == false){goto cleanup;}

  ret = netReg();
  if (ret == false) {
    net_reg_fails++;
    goto net_reg_cleanup;
  } else {
    net_reg_fails = 0;
  }

  openbearer();
  CBCread();
  CSQread();

  if (pullNetTime()) {
    Serial.print("Updating to UTC time: ");
    _dbgPrintDateTime(g_datetime);
    updateRTC(&g_datetime);
  }

  for (uint8_t i = 0; i < HTTP_UPLOAD_TRY_TIMES; i++) {
    ret = Transmit();
    if (ret == false) {
      upload_fails++;
      if (upload_fails > HTTP_FAILS_RESET_THRESHOLD) {
        upload_fails = 0;
        Serial.println(F("HTTP failure rate high. resetting"));
        hardResetSimCom();
        goto net_reg_cleanup;
      }
      continue;
    }
    break;
  }

  closeBearer();

net_reg_cleanup:
  netUnreg();

cleanup:
  serialEnd();
  tryPowerOffSimCom();
  Serial.print(F("LogToWeb() ret: "));
  Serial.println(ret);
  return ret;
}

void setup() {
  // _test();
  // while(1);


  // set up pins for sim7000
  pinMode(SIM_BUF_EN, OUTPUT);
  digitalWrite(SIM_BUF_EN, HIGH);

  pinMode(PWRKEY, OUTPUT);
  digitalWrite(PWRKEY, HIGH);

  pinMode(SIM_RESET, OUTPUT);
  digitalWrite(SIM_RESET, HIGH);

  pinMode(SIM_STATUS_BFD, INPUT);
  xDelay(1000); // we need to wait a little bit for the pin inputs to settle



  pinMode(RTC_INTERRUPT, INPUT_PULLUP);

  pinMode(Radar_VPP, OUTPUT);

  pinMode(RADAR_RST, OUTPUT);
  digitalWrite(RADAR_RST, HIGH);

  Serial.begin(BAUDRATE);

  velCom.begin(BAUDRATE);
  while (!velCom) {
  }

  digitalWrite(Radar_VPP, HIGH);
  xDelay(100);
  velread(LOGMODE);
  digitalWrite(Radar_VPP, LOW);

  preDistAverage = atoi(FDIST);
  preVelAverage = atoi(VEL);
  preAVelAverage = atoi(AVEL);

  simCom.begin(BAUDRATE);

  g_datetime.init(0, 0, 0, 0, 0, 0);

  initRTC();
  // updateRTC(&g_datetime); // set RTC to compile time

    // if (getLastStatusTimeFromSD()) { // g_datetime will be updated if successful
      // updateRTC(&g_datetime);
    // } else {
      // Serial.println(F("No valid time from SD card, RTC is set to compile time."));
    // }

  logToWeb(); // g_datetime and RTC will be updated if this process is successful


  isSDSetupDone = SDSetup();
  if (isSDSetupDone) {
    if(SD_LOG_ALL_DATA){logToSD(SD_LOG_FILE_ALL_DATA);}
    logToSD(SD_LOG_FILE_WEB_DATA);
  }

  Serial.println(F("Setup Done!"));
}

float max(float a, float b){
  return a > b ? a : b;
}

int isClose(float a, float b, float rtol, float atol){
    float tol = max(rtol * max(abs(a),abs(b)), atol); 
    return abs(a-b) <= tol;
}

int shouldLogToWeb(void){
    if (CONSTANT_LOGGING) {
        Serial.println(F("Constant logging, log now."));
        return true;
    }

    if(valid_scan_count == 0){
        Serial.println(F("No valid radar reading, log impossible values."));
        return true;
    }

    if(!isClose(velAverage, preVelAverage, THRESHOLD_PERCENT/100.0, VEL_DIFF_THRESHOLD)){
        Serial.println(F("Large change in velocity, log now."));
        return true;
    }

    if(!isClose(avelAverage, preAVelAverage, THRESHOLD_PERCENT/100.0, AVEL_DIFF_THRESHOLD)){
        Serial.println(F("Large change in velocity, log now."));
        return true;
    }

    if(!isClose(distAverage, preDistAverage, THRESHOLD_PERCENT/100.0, DIST_DIFF_THRESHOLD)){
        Serial.println(F("Large change in distance, log now."));
        return true;
    }

    // if no logging for a long time, then log
    if ((noLoggingLoopCounter + 1) * LOOP_INTERVAL_MINUTES >=
          MAX_HTTP_INTERVAL_MINUTES) {
      Serial.println(F("Long time since last log, log now."));
      return true;
    }

    return false;
}


void loop() {

  unsigned long millisTemp = timer0_millis;
  Serial.println(F("\n*** loop begin ***"));


  bool logToWebDone = false;

  if (noResetLoopsCount <= MIN_NON_RESET_LOOP_COUNT) {
    noResetLoopsCount++;
  }

  pinMode(Radar_TX, OUTPUT);
  digitalWrite(Radar_TX, HIGH);
  pinMode(Radar_RX, INPUT_PULLUP);
  velCom.begin(BAUDRATE);
  while (!velCom) {
  }

  Serial.println(F("Scanning"));
  velCom.begin(BAUDRATE);
  while (!velCom) {
  }
  digitalWrite(Radar_VPP, HIGH);
  xDelay(100);
  velread(LOGMODE);
  digitalWrite(Radar_VPP, LOW);
  Serial.println(F("Scanning done"));


  // log every scan result to SD card
  if (isSDAcessable() && SD_LOG_ALL_DATA) {
    // if SD setup is not done in setup(), try again here
    if (isSDSetupDone == false) {
      isSDSetupDone = SDSetup(); // mainly write data form titles
    }
    if (isSDSetupDone) {
      logToSD(SD_LOG_FILE_ALL_DATA);
    }
  }

  // count scan times and add up the valid radar readings
  scanCounterForHttp ++;
  if ((strlen(AVEL) > 0 || LOGMODE == 7) && strlen(FDIST) > 0) {
    int32_t vel = atol(VEL);
    int32_t avel = atol(AVEL);
    int32_t dist = atol(FDIST);
    //what about checking if the velocity is valid?
    // if (dist >= SWEEP_START && dist <= 1.1 * (SWEEP_LENGTH + SWEEP_START)) {
      velSum += vel;
      avelSum += avel;
      distSum += dist;
      valid_scan_count ++;
    // }
  }else{
    Serial.println("Missing AVEL or FDIST!");
  }

  // if it is time to upload data to web
  if (scanCounterForHttp >=
      (AVERAGING_INTERVAL_MINUTES / LOOP_INTERVAL_MINUTES)){

    getAverage();

    if (shouldLogToWeb()) {
      // convert average values to char string
      sprintf(VEL, "%" PRIi16, velAverage);
      sprintf(AVEL, "%" PRIi16, avelAverage);
      sprintf(FDIST, "%" PRIi16, distAverage);
      // if SD setup is not done, try again here
      if (isSDSetupDone == false) {
        isSDSetupDone = SDSetup(); // mainly write data form titles
      }
      // backup the data to SD card
      if (isSDSetupDone) {
        logToSD(SD_LOG_FILE_WEB_DATA);
      }
     
      if (LOG_TO_WEB ? logToWeb() : true) {
        noLoggingLoopCounter = 0;
        logToWebDone = true;
      }
      // save current average data
      preVelAverage = velAverage;
      preAVelAverage = avelAverage;
      preDistAverage = distAverage;
    } 

    // clear sums and counter no matter if log to web or not
    scanCounterForHttp = 0;
    valid_scan_count = 0;
    velSum = 0;
    avelSum = 0;
    distSum = 0;
  }

  if (logToWebDone == false) {
    if (noLoggingLoopCounter < 0xFFFFFFFD) { // avoid overflow
      noLoggingLoopCounter++;
    }
  }

  unsigned long millisDuration = timer0_millis - millisTemp;
  // if it is overflow to 0 in 50 days, it will be negative
  if (millisDuration < 0) {
    millisDuration = timer0_millis + (0xFFFFFFFF - millisTemp);
  }
  if (millisDuration < LOOP_INTERVAL_MILLIS) {
    millisTemp = LOOP_INTERVAL_MILLIS - millisDuration;
    unsigned long secs = (unsigned long)(millisTemp / 1000);
    Serial.println(F("\n*** Sleep ***"));
    xDelay(millisTemp % 1000);
    deepSleepSecs(secs);
  } else {
    Serial.println(F("\n*** No sleep, next loop ***"));
  }
}

int16_t ftoi(float value) {
    // Clamping range for int16_t
    constexpr float minInt16 = static_cast<float>(INT16_MIN);
    constexpr float maxInt16 = static_cast<float>(INT16_MAX);
    
    // Clamp the value within the range of int16_t
    if (value < minInt16) {
        return INT16_MIN;
    } else if (value > maxInt16) {
        return INT16_MAX;
    } else {
        return static_cast<int16_t>(value);
    }
}

void getAverage() {
    Serial.print(F("reps: "));
    Serial.println(valid_scan_count);
    Serial.print(F("velSum: "));
    Serial.println(velSum);
    Serial.print(F("avelSum: "));
    Serial.println(avelSum);
    Serial.print(F("distSum: "));
    Serial.println(distSum);

    if(valid_scan_count > 0){
        float va  = (float)velSum  / (float) valid_scan_count;
        float ava  = (float)avelSum  / (float) valid_scan_count;
        float da = (float)distSum / (float) valid_scan_count;
        velAverage = ftoi(va);
        avelAverage = ftoi(ava);
        distAverage = ftoi(da);
    }else{
        velAverage  = 0x7FFF;//error values which should be unreasonable in correct operation. These need to fit in the type of velAverage
        avelAverage  = 0x7FFF;//error values which should be unreasonable in correct operation. These need to fit in the type of avelAverage
        distAverage = 0x7FFF;
    }

    velSum = avelSum = distSum = 0;
}

// reads battery voltage from responce char array
void storeCBCresponse() {

  bool end = 0;
  uint8_t x = 0;
  uint8_t j = 0;

  memset(CBC, '\0', sizeof(CBC));

  // loop through reponce to extract data
  for (uint8_t i = 0; i < CHARBUFF; i++) {

    // string splitting cases
    switch (response[i]) {
      case ':':
        x = 9;
        j = 0;
        i += 2;
        break;

      case ',':
        x++;
        j = 0;
        break;

      case '\0':
        end = 1;
        break;
      case '\r':
        x++;
        j = 0;
        break;
    }
    // write to char arrays
    if (response[i] != ',') {
      switch (x) {
        case 11:
          CBC[j] = response[i];
          break;
      }
      // increment char array counter
      j++;
    }
    // break loop when end flag is high
    if (end) {
      i = CHARBUFF;
    }
  }
}


// like delay but lower power and more dodgy//
void xDelay(uint32_t tmz) {
  if (Serial) {
    Serial.flush();
  }
  uint32_t tmzslc = tmz / 64;
  // 64 chosen as ballance between power savings and time spent in full clock mode
  clock_prescale_set(clock_div_64);
  delay(tmzslc);
  clock_prescale_set(clock_div_1);

  cli();
  timer0_millis += 63 * tmzslc;
  sei();

  delay(tmz - 64 * tmzslc);
}

////power down cellular functionality////
void netUnreg() { sendATcmd(F("AT+CFUN=0"), "OK", 1000); }

//AT+COPS=? (to find all the networks) can take upwards of 20 minutes!!!
//AT+COPS=? this is the same (about 20 minutes) if search NB-IoT
//AT+COPS=? when only searching for LTE M1, it takes 90s

/* 
Typical use:

-> AT+COPS=?

<- +COPS: (2,"Telstra Mobile","Telstra","50501",7),(1,"vodafone AU","voda AU","50503",9),(1,"Optus AU","Optus","50502",9),(1,"Telstra Mobile","Telstra","50501",9),,(0,1,2,3,4),(0,1,2)

<- OK

Response format:
(stat,oper long,oper short, mccmnc, format,netact)
stat: 0, unknown; 1, available; 2, current; 3, forbidden
oper: operator name string
mccmnc: operator mccmnc
netact: 0, GSM; 1 GSM compact; 3, GSM EGPRS; 7, LTE-M1; 9, LTE-NB (NB-IoT)
*/

int _dbgPrintOperators(void) {
  Serial.print(F("Operators Length: "));
  Serial.println(operators_len);

  for (int i = 0; i < operators_len; ++i) {
    Serial.print(F("Operator "));
    Serial.print(i + 1);
    Serial.print(F(": Status - "));
    Serial.print(operators[i].status);
    Serial.print(F(", MCCMNC - "));
    Serial.print(operators[i].mccmnc);
    Serial.print(F(", Netact - "));
    Serial.print(operators[i].netact);
    Serial.print(F(", RSSI - "));
    Serial.println(operators[i].rssi);
  }
  return true;
}

int storeOperators(){
    char* start = NULL;
    char* end = NULL;
    size_t len = 0;

    operators_len = 0;

    start = strstr(response, "+COPS: ");//make sure we are reading the +COPS line
    if(start == NULL){return false;}

    for(uint8_t i = 0; i < MAX_OPERATORS; i++){
        if(start == strstr(start, ",,")){return true;} //we are at the end of the list

        start = strstr(start, "(");//start of a operators listing
        if(start == NULL){return false;}
        end   = strstr(start, "),") + 1;//end of an operator listing
        if(end == NULL){return false;}

        len = end - start;//check that the matches are of reasonable length
        if(len > CHARBUFF){return false;} 

        Operator* o = &operators[operators_len];
        uint8_t matches = sscanf(start, "(%d,%*[^,],%*[^,],\"%6[^\"]\",%d)", &o->status, &o->mccmnc, &o->netact);
        if(matches == 3){
            o->rssi = 99;//not known or not detectable
            operators_len++;
        } 
        else {return false;}

        start = end;
    }
    return false;
}

int compare_rssi(const void *a, const void *b){
    int16_t rssi_a = ((Operator*)a)->rssi;
    int16_t rssi_b = ((Operator*)b)->rssi;

    if(rssi_a == 99 && rssi_b == 99){return 0;}
    if(rssi_a == 99){return 1;}
    if(rssi_b == 99){return -1;}

    return  rssi_a > rssi_b ? -1 : 
           (rssi_b > rssi_a ?  1 : 0);

}

int netSearch(void){
    int ret = false;

    ret = sendATcmd(F("AT+COPS=?"), "OK", 1800000, 1);//I've observed this take as long as 20 minutes before so lets make the timeout 30 minutes
    if(!ret){return false;}

    storeOperators();
    _dbgPrintOperators();

    for(uint8_t i = 0; i < operators_len; i++){
        Operator* o = &operators[i];
        ret = netSelect(o->mccmnc);
        if(!ret){continue;}
        
        char mccmnc_current[7] = "\0\0\0\0\0\0\0";
        getMccmnc(mccmnc_current, sizeof(mccmnc_current));
        if(strcmp(o->mccmnc,mccmnc_current)){
          Serial.println(F("ERR: operators do not match"));
          continue;
        }

        getRSSI(&o->rssi);

    }

    qsort(operators, operators_len, sizeof(Operator), compare_rssi);  

    return true;
}


int isRegistered(const uint32_t timeout_ms = 591){
  int ret = 0;
  const uint32_t retry_delay_ms = 591;//empirically measured
  const uint32_t attempts = timeout_ms/retry_delay_ms;

  for(uint32_t i = 0; i < attempts; i++){
    ret = sendATcmd(F("AT+CREG?"), "OK", 1000, 1);
    if(ret){
      int registered = strstr(response, "+CREG: 0,5")
		       || (ACCEPT_NON_ROAMING && strstr(response, "+CREG: 0,1"));
      if(registered){
	  Serial.println(F("CREG: registered"));
	  return true;
      }
    }
    Serial.print(F("Waiting for registration, "));
    Serial.print(i);Serial.print(F("/"));Serial.println(attempts);
    xDelay(50);
  }
  Serial.println(F("CREG: not registered"));
  return false; 
}

//this supports needs to have knowledge of netact too
int netSelect(const char* mccmnc){
    int ret = 0;
    const uint32_t cops_timeout_ms = 150000;//120s specified in the AT command manual
    const uint32_t creg_timeout_ms = 150000;//this one is more how long we are willing to wait;

    if(mccmnc == NULL){
    	  Serial.println(F("Automatic network registration attempt"));
        ret = sendATcmd(F("AT+COPS=0"), "OK", cops_timeout_ms, 1);
        if(ret == 0){return false;}
    }else{
      	Serial.print(F("Attempting to register to: "));
	      Serial.println(mccmnc);
        char atcmd[64];
        snprintf(atcmd, sizeof(atcmd), "AT+COPS=4,2,%s", mccmnc);
        ret = sendATcmd(atcmd, "OK", cops_timeout_ms, 1);
        if(ret == 0){return false;}
    }
    
    ret = isRegistered(creg_timeout_ms);

    return ret;
}

////register to network////
bool netReg(void) {
  int ret;
  static char mccmnc[7] = "\0\0\0\0\0\0\0";
  static uint8_t network_searched = 0;

  if(mccmnc[0] == '\0'){//if uninitialised
    strncpy(mccmnc, MCCMNC, sizeof(mccmnc) -1 ); 
    mccmnc[sizeof(mccmnc) - 1] = '\0';
  }

  if (sendATcmd(F("AT+CFUN=1"), "OK", 10000, 2) == false) { // enable the modem
    return false;
  }

  if(ALLOW_GSM){
    sendATcmd(F("AT+CNMP=2"), "OK",
              500, 1); //enable GSM
  }else{
    sendATcmd(F("AT+CNMP=38"), "OK",
              500, 1); //disable GSM (its 2024!)
  }
            
  if(ALLOW_NBIOT){
     sendATcmd(F("AT+CMNB=3"), "OK",
               500, 1); //enable NB-IoT
  }else{
     sendATcmd(F("AT+CMNB=1"), "OK",
               500, 1); //disable NB-IoT (simbase does not offer NB-IoT)
  }

  sendATcmd(F("AT+CGDCONT=1,\"IP\",\"" APN "\""), "OK",
            2000); // set IPv4 and apn

  ret = isRegistered();
  if(ret){return ret;}

  ret = netSelect(mccmnc);
  if(ret){
      ret = getMccmnc(mccmnc, sizeof(mccmnc));
      if(ret){
        Serial.print(F("Updating network preference to: "));
        Serial.println(mccmnc);
      }
    return true; 
  }

  ret = netSelect(MCCMNC);
  if(ret){
      ret = getMccmnc(mccmnc, sizeof(mccmnc));
      if(ret){
        Serial.print(F("Updating network preference to: "));
        Serial.println(mccmnc);
      }
    return true; 
  }

  ret = netSelect(NULL);
  if(ret){
      ret = getMccmnc(mccmnc, sizeof(mccmnc));
      if(ret){
        Serial.print(F("Updating network preference to: "));
        Serial.println(mccmnc);
      }
    return true; 
  }
  
  if(!network_searched){
    Serial.println(F("Searching networks, this could take > 30 minutes."));
    netSearch();
    network_searched = true;//not we never do this again regardless of if the search was successfull
  }

  //attempt to connect to networks in order of signal strength
  for(uint8_t i = 0; i < operators_len; i++){
    Operator* o = &operators[i];
    if(o->rssi == 99){continue;}
    ret = netSelect(o->mccmnc);
    if(ret){
        ret = getMccmnc(mccmnc, sizeof(mccmnc));
        if(ret){
            Serial.print(F("Updating network preference to: "));
            Serial.println(mccmnc);
        }
        return true; 
    }
  }

  Serial.println(F("Unable to connect to network"));
  return false;
}


int getRSSI(uint8_t* rssi_p){
    int ret = 0;
    sendATcmd(F("AT+CSQ"), "OK", 1000, 1);
    char* start = strchr(response, '+');
    if(start == NULL){return false;}

    ret = sscanf(start, "+CSQ: %d,%*d", rssi_p);
    if(ret == 1){
      return true;
    }
    rssi_p = 99;
    return false;
}


int getMccmnc(char* mccmnc, size_t mccmnc_size){
    int ret = 0;
    sendATcmd(F("AT+CPSI?"), "OK", 1000, 1);
    char mcc[4] = "\0\0\0\0";
    char mnc[4] = "\0\0\0\0";
    char* start = strchr(response, '+');
    if(start == NULL){return false;}

    ret = sscanf(start, "+CPSI: %*[^,],%*[^,],%3[^-]-%3[^,],", mcc, mnc);
    if(ret == 2){
      snprintf(mccmnc, mccmnc_size, "%s%s", mcc,mnc);
      return true;
    }
    return false;
}

////sends at command, checks for reply////
bool sendATcmd(String ATcommand, char *expctAns, uint32_t timeout,
               int8_t tries) {
  uint32_t timeStart;
  bool answer;
  uint8_t a = 0;

  do {
    a++;
    Serial.println();
    Serial.println(ATcommand);

    answer = false;
    timeStart = 0;
    xDelay(100);

    while (simCom.available() > 0) {
      simCom.read(); // Clean the input buffer
    }

    simCom.println(ATcommand); // Send the AT command

    uint8_t i = 0;
    timeStart = millis();
    memset(response, '\0', CHARBUFF); // Initialize the string

    // this loop waits for the answer
    do {
      if (simCom.available() != 0) {
        response[i] = simCom.read();
        i++;
        // check if the desired answer is in the response of the module
        if (strstr(response, expctAns) != NULL) {
          answer = true;
        }
      }
      // Waits for the asnwer with time out
    } while ((answer == false) && ((millis() - timeStart) < timeout));

    Serial.print(F("response:"));
    Serial.println(response);

    if (expctAns == "\0") {
      answer = true;
    }

  } while (answer == false && a < tries);

  a = 0;
  return answer;
}

void CBCread() {
  // get GNSS data
  if (sendATcmd(F("AT+CBC"), "OK", 1000)) {

    storeCBCresponse();
  }
}

void CSQread() {
  uint8_t csq;
  getRSSI(&csq);
  memset(CSQ, '\0', sizeof(CSQ));
  sprintf(CSQ, "%" PRIu8, csq);
}

////TRANSMITS LAST GPS CORDINATES TO WEB////
bool Transmit() {

  bool ret = false;

  dataStr = "AT+HTTPPARA=\"URL\",\"http://www.bosl.com.au/IoT/";
  dataStr += SITE_DIR;
  dataStr += "/scripts/WriteMe_v2.php?SiteName=";
  dataStr += SITE_ID;
  dataStr += ".csv";
  dataStr += "&CBC=";
  dataStr += CBC;
  dataStr += "&ANGLE=";
  dataStr += ANGLE;
  dataStr += "&VEL=";
  dataStr += VEL;
  dataStr += "&PEAKS=";
  dataStr += PEAKS;
  dataStr += "&DEPTH=";
  dataStr += FDIST;
  dataStr += "&AMP=";
  dataStr += AMP;
  dataStr += "&CSQ=";
  dataStr += CSQ;
  dataStr += "&AVEL=";
  dataStr += AVEL;
  // dataStr += "&AFDIST=";
  // dataStr += AFDIST;
  dataStr += "\"";

  sendATcmd(F("AT+HTTPINIT"), "OK", 1000);
  if (strstr(response, "ERROR")) {
    return false;
  }
  sendATcmd(F("AT+HTTPPARA=\"CID\",1"), "OK", 1000);
  if (strstr(response, "ERROR")) {
    return false;
  }

  sendATcmd(dataStr, "OK", 2000);  // check what the HTTPTIMEOUT is

  if (sendATcmd(F("AT+HTTPACTION=0"), "OK", 5000) == false) {
    return false;
  }

  ret = sendATcmd(F("AT+HTTPSTATUS?"), "+HTTPACTION: 0,200", 1000, 1);
  if (strstr(response, "ERROR")) {
    return false; // if the response is ERROR, no need to continue
  }

  // await the "200" response a bit longer to ensure http uploading is done
  if (ret == false) {
    sendATcmd(F("AT+HTTPSTATUS?"), "+HTTPACTION: 0,200", 1000, 15); // TODO: what is the best timeout?
  }
  
  sendATcmd(F("AT+HTTPSTATUS?"), "+HTTPSTATUS: GET,0,0,0", 1000, 2);
  sendATcmd(F("AT+HTTPTERM"), "OK", 1000, 3);
  return true;
}

void deepSleepSecs(int32_t sec) {
  // turn adc off
  ADCSRA &= ~(1 << ADEN);

  while (sec >= 4) {
    sec -= 4;
    // wdt wakeup timer
    wdt_enable(WDTO_4S);
    WDTCSR |= (1 << WDIE);

    // sleep
    sleepMode(SLEEP_POWER_DOWN);
    sleep();
    noSleep();
    timer0_millis += 4000;
  }

  while (sec >= 1) {
    sec -= 1;
    // wdt wakeup timer
    wdt_enable(WDTO_1S);
    WDTCSR |= (1 << WDIE);

    // sleep
    sleepMode(SLEEP_POWER_DOWN);
    sleep();
    noSleep();
    timer0_millis += 1000;
  }

  // adc on
  ADCSRA |= (1 << ADEN);
}

ISR(WDT_vect) {
  // WDIE & WDIF is cleared in hardware upon entering this ISR
  wdt_disable();
}
